@@ -198,6 +198,25 @@ module Path = struct
     | `ClassType (p, name) -> `ClassType (resolved_parent map p, name)
     | `Substituted s -> resolved_type map s
 
+  and resolved_value map (`Value (p, name) : Cpath.Resolved.value) :
+      Odoc_model.Paths.Path.Resolved.Value.t =
+    `Value (resolved_parent map p, name)
+
+  and resolved_datatype map (p : Cpath.Resolved.datatype) :
+      Odoc_model.Paths.Path.Resolved.DataType.t =
+    match p with
+    | `Gpath y -> y
+    | `Local id -> `Identifier (Component.TypeMap.find id map.type_)
+    | `CanonicalDataType (t1, t2) ->
+        `CanonicalDataType (resolved_datatype map t1, t2)
+    | `Type (p, name) -> `Type (resolved_parent map p, name)
+    | `Substituted s -> resolved_datatype map s
+
+  and resolved_constructor map
+      (`Constructor (p, name) : Cpath.Resolved.constructor) :
+      Odoc_model.Paths.Path.Resolved.Constructor.t =
+    `Constructor (resolved_datatype map p, name)
+
   and resolved_class_type map (p : Cpath.Resolved.class_type) :
       Odoc_model.Paths.Path.Resolved.ClassType.t =
     match p with
@@ -478,6 +497,7 @@ and class_ map parent id c =
   in
   {
     id = identifier;
+    locs = c.locs;
     doc = docs (parent :> Identifier.LabelParent.t) c.doc;
     virtual_ = c.virtual_;
     params = c.params;
@@ -493,7 +513,7 @@ and class_decl map parent c =
   | Arrow (lbl, t, d) ->
       Arrow
         ( lbl,
-          type_expr map (parent :> Identifier.Parent.t) t,
+          type_expr map (parent :> Identifier.LabelParent.t) t,
           class_decl map parent d )
 
 and class_type_expr map parent c =
@@ -501,7 +521,7 @@ and class_type_expr map parent c =
   | Component.ClassType.Constr (p, ts) ->
       Constr
         ( Path.class_type map p,
-          List.rev_map (type_expr map (parent :> Identifier.Parent.t)) ts
+          List.rev_map (type_expr map (parent :> Identifier.LabelParent.t)) ts
           |> List.rev )
   | Signature s -> Signature (class_signature map parent s)
 
@@ -515,6 +535,7 @@ and class_type map parent id c =
   in
   {
     Odoc_model.Lang.ClassType.id = identifier;
+    locs = c.locs;
     doc = docs (parent :> Identifier.LabelParent.t) c.doc;
     virtual_ = c.virtual_;
     params = c.params;
@@ -527,7 +548,7 @@ and class_type map parent id c =
 
 and class_signature map parent sg =
   let open Component.ClassSignature in
-  let pparent = (parent :> Identifier.Parent.t) in
+  let pparent = (parent :> Identifier.LabelParent.t) in
   let items =
     List.rev_map
       (function
@@ -552,7 +573,7 @@ and method_ map parent id m =
     doc = docs (parent :> Identifier.LabelParent.t) m.doc;
     private_ = m.private_;
     virtual_ = m.virtual_;
-    type_ = type_expr map (parent :> Identifier.Parent.t) m.type_;
+    type_ = type_expr map (parent :> Identifier.LabelParent.t) m.type_;
   }
 
 and instance_variable map parent id i =
@@ -566,7 +587,7 @@ and instance_variable map parent id i =
     doc = docs (parent :> Identifier.LabelParent.t) i.doc;
     mutable_ = i.mutable_;
     virtual_ = i.virtual_;
-    type_ = type_expr map (parent :> Identifier.Parent.t) i.type_;
+    type_ = type_expr map (parent :> Identifier.LabelParent.t) i.type_;
   }
 
 and class_constraint map parent cst =
@@ -663,8 +684,9 @@ and value_ map parent id v =
   let identifier = Identifier.Mk.value (parent, typed_name) in
   {
     id = identifier;
+    locs = v.locs;
     doc = docs (parent :> Identifier.LabelParent.t) v.doc;
-    type_ = type_expr map (parent :> Identifier.Parent.t) v.type_;
+    type_ = type_expr map (parent :> Identifier.LabelParent.t) v.type_;
     value = v.value;
   }
 
@@ -686,10 +708,13 @@ and extension_constructor map parent c =
   in
   {
     id = identifier;
+    locs = c.locs;
     doc = docs (parent :> Identifier.LabelParent.t) c.doc;
     args =
-      type_decl_constructor_argument map (parent :> Identifier.Parent.t) c.args;
-    res = Opt.map (type_expr map (parent :> Identifier.Parent.t)) c.res;
+      type_decl_constructor_argument map
+        (parent :> Identifier.FieldParent.t)
+        c.args;
+    res = Opt.map (type_expr map (parent :> Identifier.LabelParent.t)) c.res;
   }
 
 and module_ map parent id m =
@@ -702,6 +727,7 @@ and module_ map parent id m =
     let map = { map with shadowed = empty_shadow } in
     {
       Odoc_model.Lang.Module.id;
+      locs = m.locs;
       doc = docs (parent :> Identifier.LabelParent.t) m.doc;
       type_ = module_decl map identifier m.type_;
       canonical = m.canonical;
@@ -743,11 +769,11 @@ and mty_substitution map identifier = function
   | TypeEq (frag, eqn) ->
       TypeEq
         ( Path.type_fragment map frag,
-          type_decl_equation map (identifier :> Identifier.Parent.t) eqn )
+          type_decl_equation map (identifier :> Identifier.FieldParent.t) eqn )
   | TypeSubst (frag, eqn) ->
       TypeSubst
         ( Path.type_fragment map frag,
-          type_decl_equation map (identifier :> Identifier.Parent.t) eqn )
+          type_decl_equation map (identifier :> Identifier.FieldParent.t) eqn )
   | ModuleTypeEq (frag, eqn) ->
       ModuleTypeEq
         (Path.module_type_fragment map frag, module_type_expr map identifier eqn)
@@ -840,6 +866,7 @@ and module_type :
   let map = { map with shadowed = empty_shadow } in
   {
     Odoc_model.Lang.ModuleType.id = identifier;
+    locs = mty.locs;
     doc = docs (parent :> Identifier.LabelParent.t) mty.doc;
     canonical = mty.canonical;
     expr = Opt.map (module_type_expr map sig_id) mty.expr;
@@ -863,17 +890,20 @@ and module_type_substitution :
 
 and type_decl_constructor_argument :
     maps ->
-    Paths.Identifier.Parent.t ->
+    Paths.Identifier.FieldParent.t ->
     Component.TypeDecl.Constructor.argument ->
     Odoc_model.Lang.TypeDecl.Constructor.argument =
  fun map parent a ->
   match a with
-  | Tuple ls -> Tuple (List.map (type_expr map parent) ls)
-  | Record fs -> Record (List.map (type_decl_field map parent) fs)
+  | Tuple ls ->
+      Tuple (List.map (type_expr map (parent :> Identifier.LabelParent.t)) ls)
+  | Record fs ->
+      Record
+        (List.map (type_decl_field map (parent :> Identifier.FieldParent.t)) fs)
 
 and type_decl_field :
     maps ->
-    Identifier.Parent.t ->
+    Identifier.FieldParent.t ->
     Component.TypeDecl.Field.t ->
     Odoc_model.Lang.TypeDecl.Field.t =
  fun map parent f ->
@@ -882,12 +912,13 @@ and type_decl_field :
     id = identifier;
     doc = docs (parent :> Identifier.LabelParent.t) f.doc;
     mutable_ = f.mutable_;
-    type_ = type_expr map parent f.type_;
+    type_ = type_expr map (parent :> Identifier.LabelParent.t) f.type_;
   }
 
-and type_decl_equation map (parent : Identifier.Parent.t)
+and type_decl_equation map (parent : Identifier.FieldParent.t)
     (eqn : Component.TypeDecl.Equation.t) : Odoc_model.Lang.TypeDecl.Equation.t
     =
+  let parent = (parent :> Identifier.LabelParent.t) in
   {
     params = eqn.params;
     private_ = eqn.private_;
@@ -903,7 +934,9 @@ and type_decl map parent id (t : Component.TypeDecl.t) :
   let identifier = Component.TypeMap.find id map.type_ in
   {
     id = identifier;
-    equation = type_decl_equation map (parent :> Identifier.Parent.t) t.equation;
+    locs = t.locs;
+    equation =
+      type_decl_equation map (parent :> Identifier.FieldParent.t) t.equation;
     doc = docs (parent :> Identifier.LabelParent.t) t.doc;
     canonical = t.canonical;
     representation =
@@ -918,29 +951,29 @@ and type_decl_representation map id (t : Component.TypeDecl.Representation.t) :
   | Record fs ->
       Record
         (List.map
-           (type_decl_field map (id :> Odoc_model.Paths.Identifier.Parent.t))
+           (type_decl_field map
+              (id :> Odoc_model.Paths.Identifier.FieldParent.t))
            fs)
 
 and type_decl_constructor :
     maps ->
-    Odoc_model.Paths.Identifier.Type.t ->
+    Odoc_model.Paths.Identifier.DataType.t ->
     Component.TypeDecl.Constructor.t ->
     Odoc_model.Lang.TypeDecl.Constructor.t =
  fun map id t ->
   let identifier =
     Identifier.Mk.constructor (id, ConstructorName.make_std t.name)
   in
+  let parent = (id :> Identifier.LabelParent.t) in
   {
     id = identifier;
-    doc = docs (id :> Identifier.LabelParent.t) t.doc;
+    doc = docs parent t.doc;
     args =
-      type_decl_constructor_argument map
-        (id :> Odoc_model.Paths.Identifier.Parent.t)
-        t.args;
-    res = Opt.map (type_expr map (id :> Identifier.Parent.t)) t.res;
+      type_decl_constructor_argument map (id :> Identifier.FieldParent.t) t.args;
+    res = Opt.map (type_expr map parent) t.res;
   }
 
-and type_expr_package map parent t =
+and type_expr_package map (parent : Identifier.LabelParent.t) t =
   {
     Lang.TypeExpr.Package.path =
       Path.module_type map t.Component.TypeExpr.Package.path;
@@ -951,8 +984,8 @@ and type_expr_package map parent t =
         t.substitutions;
   }
 
-and type_expr map (parent : Identifier.Parent.t) (t : Component.TypeExpr.t) :
-    Odoc_model.Lang.TypeExpr.t =
+and type_expr map (parent : Identifier.LabelParent.t) (t : Component.TypeExpr.t)
+    : Odoc_model.Lang.TypeExpr.t =
   try
     match t with
     | Var s -> Var s
@@ -984,7 +1017,7 @@ and type_expr_polyvar map parent v =
         c.Component.TypeExpr.Polymorphic_variant.Constructor.name;
       constant = c.constant;
       arguments = List.map (type_expr map parent) c.arguments;
-      doc = docs (parent :> Identifier.LabelParent.t) c.doc;
+      doc = docs parent c.doc;
     }
   in
   let element = function
@@ -1025,10 +1058,13 @@ and exception_ map parent id (e : Component.Exception.t) :
   in
   {
     id = identifier;
+    locs = e.locs;
     doc = docs (parent :> Identifier.LabelParent.t) e.doc;
     args =
-      type_decl_constructor_argument map (parent :> Identifier.Parent.t) e.args;
-    res = Opt.map (type_expr map (parent :> Identifier.Parent.t)) e.res;
+      type_decl_constructor_argument map
+        (parent :> Identifier.FieldParent.t)
+        e.args;
+    res = Opt.map (type_expr map (parent :> Identifier.LabelParent.t)) e.res;
   }
 
 and block_element parent
